import pytest
from pytest_mock import MockerFixture

from workspaces.schema import WorkflowEventType
from wf_monitor.monitor import WorkflowMonitor
from channels.amqp_helpers import Channel, workflow_events

log_path = 'logs/condor.log'
test_monitor = WorkflowMonitor(log_path)
WORKFLOW_STATUS_EXCH = 'workspaces.workflow-service.workflow-status'


def test_read_log():
    """
    Tests whether or not the example log is being correctly read by checking for strings known to
    exist in the example log file
    """
    test_strs = [
        'Image size of job updated: 432',
        '432  -  ResidentSetSize of job (KB)',
        '(1) Normal termination (return value 0)',
        '000 (3983.000.000) 08/26 11:06:06 Job submitted from host: '
        '<10.64.1.178:9618?addrs=10.64.1.178-9618&noUDP&sock=4050180_7e37_3>'
    ]
    for s in test_strs:
        assert s in test_monitor.log


def test_read_log_timeout():
    """
    Tests the timeout functionality of WorkflowMonitor.read_htcondor_log()
    """
    with pytest.raises(SystemExit) as sys_ex:
        WorkflowMonitor('logs/file-that-does-not-exist.txt', 1)
    assert sys_ex.type == SystemExit
    assert sys_ex.value.code == -1


def test_parse_log():
    """
    Tests that the example log file is correctly parsed by checking for the known order of event
    types
    """
    test_event_types = [
        WorkflowEventType.SUBMITTED,
        WorkflowEventType.EXECUTING,
        WorkflowEventType.OTHER,
        WorkflowEventType.TERMINATED,
        WorkflowEventType.SUBMITTED,
        WorkflowEventType.EXECUTING,
        WorkflowEventType.OTHER,
        WorkflowEventType.TERMINATED
    ]
    for e, e_type in zip(test_monitor.events, test_event_types):
        assert e.type == e_type


def test_parse_log_error():
    """
    Tests that WorkflowMonitor.parse_log() correctly raises a ValueError when a badly formatted
    log is parsed
    """
    with pytest.raises(ValueError) as val_err:
        WorkflowMonitor('logs/test.log')
    assert val_err.type == ValueError


def test_monitor_events(mocker: MockerFixture):
    """
    Test that mocks the core wf_monitor functionality of creating a connection to the RabbitMQ
    server, configuring a channel with that connection, and emitting events through that channel
    :param mocker: Pytest fixture from pytest_mock that wraps the real function and mimics that
    its been called as normal; allows for testing interaction with the server without actually
    requiring a connection to the server
    """
    mock_list = [
        'pika.adapters.blocking_connection.BlockingChannel.basic_publish',
        'channels.amqp_helpers.Channel.connect',
        'channels.amqp_helpers.Channel.send',
    ]
    [mocker.patch(mock) for mock in mock_list]

    with workflow_events as wfe:
        for event in WorkflowMonitor(log_path).events:
            wfe.send(event, WORKFLOW_STATUS_EXCH)

    assert wfe.send.call_count == 8
