"""
Interfaces used by the Workspace system.
"""

import inspect
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from threading import Thread
from typing import Dict, List, Optional, Type

ProductLocator = str
CapabilityName = str


@dataclass
class Capability(ABC):
    """
    A capability
    """
    name: CapabilityName
    max_jobs: int

    def create_request(self, locators: List[ProductLocator]):
        """
        Create a capability request for this capability

        :param locators:  product locators for the new request
        :return:          a capability request
        """
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


class ProductService(ABC):
    """
    Locate products and realize them on disk (haha).
    """
    @abstractmethod
    def locate_product(self, product_locator: ProductLocator) -> Path:
        """
        Locates a given product and produces a file path to it.
        :param product_locator:   the locator to this product
        :return:                  a path to this product
        """
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


@dataclass
class CapabilityRequest:
    """
    A particular capability request
    """
    # From wksp0
    capability: Capability
    locators: List[ProductLocator]
    id: Optional[int]
    parameters: List["Parameter"]
    files: List[Path]
    # From class diagram
    # FIXME: Will versions be str or a class of their own?
    versions: List[str]
    tickets: List[Ticket]
    queue: "CapabilityQueue"

    @property
    def last_parameter(self) -> "Parameter":
        return self.parameters[-1]


@dataclass
class CapabilityInfo(ABC):
    """
    Interface to stored capability information.
    """
    # TODO: Implement these two types
    matrix: CapabilityMatrix
    proj_settings: ProjectSettings

    @abstractmethod
    def lookup_capability(self, capability_name: str) -> Capability:
        """
        Look up the definition of a capability.

        :param capability_name:  the name of the capability to find
        :return: a capability
        """
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')

    @abstractmethod
    def lookup_capability_request(self, capability_request_id: int) -> CapabilityRequest:
        """
        Look up a particular request
        :param capability_request_id:  the request identifier
        :return: a capability request
        """
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')

    @abstractmethod
    def save_request(self, request: CapabilityRequest) -> int:
        """
        Save a capability request and return an integer identifier for it.

        :param request:  the request to save
        :return:         the request identifier
        """
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


class QueueRunner(Thread, ABC):
    pass


class CapabilityQueue(ABC):
    """
    Holds capability requests until they can be executed.
    """
    @abstractmethod
    def enqueue(self, request: CapabilityRequest):
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


@dataclass
class CapabilityService(ABC):
    """
    The capability service: clients access this to request capability runs
    """
    # TODO: Implement EventTranslator and CapabilityFactory
    event_translator: EventTranslator
    queue: CapabilityQueue
    info: CapabilityInfo
    factory: CapabilityFactory

    # TODO: Implement two services
    workflow: "WorkflowServiceIF"
    estimation_service: EstimationService
    metrics_service: MetricsService

    @abstractmethod
    def send_request(self, name: CapabilityName, locators: List[ProductLocator]) -> CapabilityRequest:
        """
        Start a capability request with the given capability name and product locators.

        :param name:      the capability name to look things up with
        :param locators:  the products to start the capability with
        :return:          a new capability request
        """
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')

    @abstractmethod
    def execute(self, request: CapabilityRequest) -> None:
        """
        Begin executing a capability request

        :param request:  the request to execute
        :return: None
        """
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


class ArchiveService(ABC):
    """
    Abstracts services that are needed from the archive system.
    """
    @abstractmethod
    def lookup_product(self, locator: ProductLocator) -> ScienceProduct:
        """
        Look up a science product by its locator
        :param locator:  science product locator for this product
        :return:         science product
        """
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


FieldName = str
FieldLabel = str


class Parameter(ABC):
    """
    Abstracts parameters needed for running capabilities.
    """
    @staticmethod
    @abstractmethod
    def fields() -> Dict[FieldName, FieldLabel]:
        raise NotImplementedError(f'Parameter.{inspect.stack()[0][3]}')

    @abstractmethod
    def json(self) -> Dict[str, str]:
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')

    @abstractmethod
    def load(self, json: Dict[str, str]) -> None:
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


@dataclass
class CapabilityStep(ABC):
    """
    A step in a capability sequence
    """
    next: CapabilityStep

    @abstractmethod
    def execute_against(self, request: CapabilityRequest, responder: "CapabilityEngineResponder"):
        """
        Execute this capability step
        :return: None
        """
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


@dataclass
class CapabilityExecution(ABC):
    request: WorkflowRequest
    # FIXME: Will this be a string?
    version: str
    current_step: CapabilityStep
    sequence: CapabilitySequence


class CapabilityEngineResponder(ABC):
    """
    Abstracts the callbacks for a capability engine
    """
    @abstractmethod
    def await_parameter(self, step: CapabilityStep, parameter_type: Type[Parameter]) -> Parameter:
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')

    @abstractmethod
    def await_product(self, step: CapabilityStep, product_locator: ProductLocator):
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')

    @abstractmethod
    def prepare_and_run_workflow(self, step: CapabilityStep, name: str, param: Parameter, files: List[Path]):
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


@dataclass
class CapabilitySequence(ABC):
    """
    Represents the sequence of events in a capability.
    """
    steps: List[CapabilityStep]


class CapabilityEngine(ABC):
    """
    Executes a capability.
    """
    @abstractmethod
    def execute(self, request):
        pass


class AwaitProduct(CapabilityStep, ABC):
    """
    Wait for a product to become available.
    """
    # FIXME: Does this need to be a class attribute?
    product: ProductLocator

    def __init__(self, product: Optional[ProductLocator] = None):
        self.product = product

    def execute_against(self, request: CapabilityRequest, responder: CapabilityEngineResponder):
        # if we have a product, await it
        if self.product:
            request.files.append(responder.await_product(self, self.product))

        # if we do not, await the locators on the request itself
        else:
            for locator in request.locators:
                request.files.append(responder.await_product(self, locator))


class AwaitParameter(CapabilityStep, ABC):
    """
    Wait for a certain parameter to arrive (probably from the UI).
    """
    # FIXME: Does this need to be a class attribute?
    parameter_type: Type[Parameter]

    def __init__(self, parameter_type: Type[Parameter]):
        self.parameter_type = parameter_type

    def execute_against(self, request: CapabilityRequest, responder: CapabilityEngineResponder):
        request.parameters.append(responder.await_parameter(self, self.parameter_type))


class PrepareAndRunWorkflow(CapabilityStep, ABC):
    """
    Render templates and execute a workflow, awaiting its completion.
    """
    # FIXME: Does this need to be a class attribute?
    workflow_name: str

    def __init__(self, workflow_name: str):
        self.workflow_name = workflow_name

    def execute_against(self, request: CapabilityRequest, responder: CapabilityEngineResponder):
        responder.prepare_and_run_workflow(self, self.workflow_name, request.last_parameter, request.files)


class WorkflowServiceIF(ABC):
    """
    Executes workflows; should be a freestanding service.
    """
    @abstractmethod
    def execute(self, workflow_name: str, argument: Dict, files: List[Path]):
        """
        Execute this workflow against these files.

        :param workflow_name:  name of the workflow to run
        :param argument:       extra argument (a JSON object)
        :param files:          some extra files the workflow should consider
        :return:               a stream of events from this workflow
        """
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


@dataclass
class WorkflowInfoIF(ABC):
    """
    Holds information about workflows.
    """
    # TODO: Implement WorkflowDefinition
    definition: WorkflowDefinition
    request: "WorkflowRequest"

    @abstractmethod
    def lookup_workflow_definition(self, name: str) -> 'Workflow':
        """
        Look up the workflow with this name.

        :param name:  Workflow name
        :return:      Workflow instance
        """
        raise NotImplementedError(f'{self.__class__.__name__}.{inspect.stack()[0][3]}')


class WorkflowDefinition(ABC):
    pass


@dataclass
class WorkflowRequest(ABC):
    workflow_definition: WorkflowDefinition
    name: str
    argument: str  # JSON-formatted
    files: List[Path]
    job_id: str